@@ -248,7 +248,7 @@ TEST_F(TestExecutor, spin_all_invalid_duration) {
 
   RCLCPP_EXPECT_THROW_EQ(
     dummy.spin_all(std::chrono::nanoseconds(-1)),
-    std::invalid_argument("max_duration must be positive"));
+    std::invalid_argument("max_duration must be greater than or equal to 0"));
 }
 
 TEST_F(TestExecutor, spin_some_in_spin_some) {
@@ -546,6 +546,40 @@ TEST_F(TestExecutor, spin_node_once_node) {
   EXPECT_TRUE(spin_called);
 }
 
+TEST_F(TestExecutor, spin_node_all_base_interface) {
+  DummyExecutor dummy;
+  auto node = std::make_shared<rclcpp::Node>("node", "ns");
+  bool spin_called = false;
+  auto timer =
+    node->create_wall_timer(
+    std::chrono::milliseconds(1), [&]() {
+      spin_called = true;
+    });
+
+  // Wait for the wall timer to have expired.
+  std::this_thread::sleep_for(std::chrono::milliseconds(50));
+  EXPECT_FALSE(spin_called);
+  dummy.spin_node_all(node->get_node_base_interface(), std::chrono::milliseconds(50));
+  EXPECT_TRUE(spin_called);
+}
+
+TEST_F(TestExecutor, spin_node_all_node) {
+  DummyExecutor dummy;
+  auto node = std::make_shared<rclcpp::Node>("node", "ns");
+  bool spin_called = false;
+  auto timer =
+    node->create_wall_timer(
+    std::chrono::milliseconds(1), [&]() {
+      spin_called = true;
+    });
+
+  // Wait for the wall timer to have expired.
+  std::this_thread::sleep_for(std::chrono::milliseconds(50));
+  EXPECT_FALSE(spin_called);
+  dummy.spin_node_all(node, std::chrono::milliseconds(50));
+  EXPECT_TRUE(spin_called);
+}
+
 TEST_F(TestExecutor, spin_until_future_complete_future_already_complete) {
   DummyExecutor dummy;
   auto node = std::make_shared<rclcpp::Node>("node", "ns");
@@ -556,3 +590,24 @@ TEST_F(TestExecutor, spin_until_future_complete_future_already_complete) {
     rclcpp::FutureReturnCode::SUCCESS,
     dummy.spin_until_future_complete(future, std::chrono::milliseconds(1)));
 }
+
+TEST_F(TestExecutor, is_spinning) {
+  DummyExecutor dummy;
+  ASSERT_FALSE(dummy.is_spinning());
+
+  auto node = std::make_shared<rclcpp::Node>("node", "ns");
+  bool timer_called = false;
+  auto timer =
+    node->create_wall_timer(
+    std::chrono::milliseconds(1), [&]() {
+      timer_called = true;
+      EXPECT_TRUE(dummy.is_spinning());
+    });
+
+  dummy.add_node(node);
+  // Wait for the wall timer to have expired.
+  std::this_thread::sleep_for(std::chrono::milliseconds(50));
+  dummy.spin_some(std::chrono::milliseconds(1));
+
+  ASSERT_TRUE(timer_called);
+}
